"""AbstractDownsampler interface-class, subclassed by concrete downsamplers."""

__author__ = "Jeroen Van Der Donckt"

import re
import warnings
from abc import ABC, abstractmethod
from copy import deepcopy
from types import ModuleType
from typing import Callable, List, Optional, Tuple, Union

import numpy as np


class AbstractDownsampler(ABC):
    """AbstractDownsampler interface-class, subclassed by concrete downsamplers."""

    def __init__(
        self,
        check_contiguous: bool = True,
        x_dtype_regex_list: Optional[List[str]] = None,
        y_dtype_regex_list: Optional[List[str]] = None,
    ):
        self.check_contiguous = check_contiguous
        self.x_dtype_regex_list = x_dtype_regex_list
        self.y_dtype_regex_list = y_dtype_regex_list

    def _check_contiguous(self, arr: np.ndarray, y: bool = True):
        # necessary for rust downsamplers as they don't support non-contiguous arrays
        # (we call .as_slice().unwrap() on the array) in the lib.rs file
        # which will panic if the array is not contiguous
        if not self.check_contiguous:
            return

        if arr.flags["C_CONTIGUOUS"]:
            return

        raise ValueError(f"{'y' if y else 'x'} array must be contiguous.")

    def _supports_dtype(self, arr: np.ndarray, y: bool = True):
        dtype_regex_list = self.y_dtype_regex_list if y else self.x_dtype_regex_list
        # base case
        if dtype_regex_list is None:
            return

        for dtype_regex_str in dtype_regex_list:
            m = re.compile(dtype_regex_str).match(str(arr.dtype))
            if m is not None:  # a match is found
                return
        raise ValueError(
            f"{arr.dtype} doesn't match with any regex in {dtype_regex_list} "
            f"for the {'y' if y else 'x'}-data"
        )

    @staticmethod
    def _check_valid_downsample_args(
        *args,
    ) -> Tuple[Union[np.ndarray, None], np.ndarray]:
        if len(args) == 2:
            x, y = args
        elif len(args) == 1:
            x, y = None, args[0]
        else:
            raise ValueError(
                "downsample() takes 1 or 2 positional arguments but "
                f"{len(args)} were given"
            )

        if x is not None and not isinstance(x, np.ndarray):
            x = np.array(x)
        if not isinstance(y, np.ndarray):
            y = np.array(y)

        # y must be 1D array
        if y.ndim != 1:
            raise ValueError("y must be 1D array")
        # x must be 1D array with same length as y or None
        if x is not None:
            if x.ndim != 1:
                raise ValueError("x must be 1D array")
            if len(x) != len(y):
                raise ValueError("x and y must have the same length")

        return x, y

    @staticmethod
    def _check_valid_n_out(n_out: int):
        if n_out <= 0:
            raise ValueError("n_out must be greater than 0")

    @abstractmethod
    def _downsample(
        self, x: Union[np.ndarray, None], y: np.ndarray, n_out: int, **kwargs
    ) -> np.ndarray:
        """Downsample the data in x and y.

        Returns
        -------
        np.ndarray
            The selected indices.
        """
        raise NotImplementedError

    def downsample(self, *args, n_out: int, **kwargs):  # x and y are optional
        """Downsample y (and x).

        Call signatures::
            downsample([x], y, n_out, **kwargs)


        Parameters
        ----------
        x, y : array-like
            The horizontal / vertical coordinates of the data points.
            *x* values are optional.
            These parameters should be 1D arrays.
            These arguments cannot be passed as keywords.
        n_out : int
            The number of points to keep.
        **kwargs
            Additional keyword arguments are passed to the downsampler.

        Returns
        -------
        np.ndarray
            The selected indices.
        """
        self._check_valid_n_out(n_out)
        x, y = self._check_valid_downsample_args(*args)
        self._supports_dtype(y, y=True)
        self._check_contiguous(y, y=True)
        if x is not None:
            self._supports_dtype(x, y=False)
            self._check_contiguous(x, y=False)
        return self._downsample(x, y, n_out, **kwargs)


# ------------------- Rust Downsample Interface -------------------
DOWNSAMPLE_F = "downsample"


# the following dtypes are supported by the rust downsamplers (x and y)
_rust_dtypes = [
    "float32",
    "float64",
    "uint16",
    "uint32",
    "uint64",
    "int16",
    "int32",
    "int64",
    "datetime64",
    "timedelta64",
]
# <= 8-bit x-dtypes are not supported as the range of the values is too small to require
# downsampling
_y_rust_dtypes = _rust_dtypes + ["float16", "int8", "uint8", "bool"]


class AbstractRustDownsampler(AbstractDownsampler, ABC):
    """RustDownsampler interface-class, subclassed by concrete downsamplers."""

    def __init__(self):
        super().__init__(True, _rust_dtypes, _y_rust_dtypes)  # same for x and y

    @property
    def _downsample_func_prefix(self) -> str:
        """The prefix of the downsample functions in the rust module."""
        return DOWNSAMPLE_F

    @property
    def rust_mod(self) -> ModuleType:
        """The compiled Rust module for the current downsampler."""
        raise NotImplementedError

    @property
    def mod_single_core(self) -> ModuleType:
        """Get the single-core Rust module.

        Returns
        -------
        ModuleType
            If SIMD compiled module is available, that one is returned. Otherwise, the
            scalar compiled module is returned.
        """
        return self.rust_mod.sequential

    @property
    def mod_multi_core(self) -> Union[ModuleType, None]:
        """Get the multi-core Rust module.

        Returns
        -------
        ModuleType or None
            If SIMD parallel compiled module is available, that one is returned.
            Otherwise, the scalar parallel compiled module is returned.
            If no parallel compiled module is available, None is returned.
        """
        if hasattr(self.rust_mod, "parallel"):
            # use SIMD implementation if available
            return self.rust_mod.parallel
        return None  # no parallel compiled module available

    @staticmethod
    def _view_x(x: np.ndarray) -> np.ndarray:
        """View the x-data as different dtype (if necessary)."""
        if np.issubdtype(x.dtype, np.datetime64):
            # datetime64 is viewed as int64
            return x.view(dtype=np.int64)
        elif np.issubdtype(x.dtype, np.timedelta64):
            # timedelta64 is viewed as int64
            return x.view(dtype=np.int64)
        return x

    @staticmethod
    def _view_y(y: np.ndarray) -> np.ndarray:
        """View the y-data as different dtype (if necessary)."""
        if y.dtype == "bool":
            # bool is viewed as int8
            return y.view(dtype=np.int8)
        elif np.issubdtype(y.dtype, np.datetime64):
            # datetime64 is viewed as int64
            return y.view(dtype=np.int64)
        elif np.issubdtype(y.dtype, np.timedelta64):
            # timedelta64 is viewed as int64
            return y.view(dtype=np.int64)
        return y

    def _switch_mod_with_y(
        self, y_dtype: np.dtype, mod: ModuleType, downsample_func: Optional[str] = None
    ) -> Callable:
        """Select the appropriate function from the rust module for the y-data.

        Assumes equal binning (when no data for x is passed -> only this function is
        executed).
        Equidistant binning is  utilized when a `downsample_func` is passed from the
        `_switch_mod_with_x_and_y` method (since the x-data is considered in the
        downsampling).

        Parameters
        ----------
        y_dtype : np.dtype
            The dtype of the y-data
        mod : ModuleType
            The module to select the appropriate function from
        downsample_func : str, optional
            The name of the function to use, by default DOWNSAMPLE_FUNC.
            This argument is passed from the `_switch_mod_with_x_and_y` method when
            the x-data is considered in the downsampling.
        """
        if downsample_func is None:
            downsample_func = self._downsample_func_prefix
        # FLOATS
        if np.issubdtype(y_dtype, np.floating):
            if y_dtype == np.float16:
                return getattr(mod, downsample_func + "_f16")
            elif y_dtype == np.float32:
                return getattr(mod, downsample_func + "_f32")
            elif y_dtype == np.float64:
                return getattr(mod, downsample_func + "_f64")
        # UINTS
        elif np.issubdtype(y_dtype, np.unsignedinteger):
            if y_dtype == np.uint8:
                return getattr(mod, downsample_func + "_u8")
            elif y_dtype == np.uint16:
                return getattr(mod, downsample_func + "_u16")
            elif y_dtype == np.uint32:
                return getattr(mod, downsample_func + "_u32")
            elif y_dtype == np.uint64:
                return getattr(mod, downsample_func + "_u64")
        # INTS (need to be last because uint is subdtype of int)
        elif np.issubdtype(y_dtype, np.integer):
            if y_dtype == np.int8:
                return getattr(mod, downsample_func + "_i8")
            elif y_dtype == np.int16:
                return getattr(mod, downsample_func + "_i16")
            elif y_dtype == np.int32:
                return getattr(mod, downsample_func + "_i32")
            elif y_dtype == np.int64:
                return getattr(mod, downsample_func + "_i64")
        # DATETIME -> i64 (datetime64 is viewed as int64)
        # TIMEDELTA -> i64 (timedelta64 is viewed as int64)
        # BOOLS -> int8 (bool is viewed as int8)
        raise ValueError(f"Unsupported data type (for y): {y_dtype}")

    def _switch_mod_with_x_and_y(
        self,  # necessary to access the class its _switch_mod_with_y method
        x_dtype: np.dtype,
        y_dtype: np.dtype,
        mod: ModuleType,
        downsample_func: Optional[str] = None,
    ) -> Callable:
        """The x-data is considered in the downsampling

        Assumes equal binning.

        Parameters
        ----------
        x_dtype : np.dtype
            The dtype of the x-data
        y_dtype : np.dtype
            The dtype of the y-data
        mod : ModuleType
            The module to select the appropriate function from
        downsample_func : str, optional
            The name of the function to use, by default DOWNSAMPLE_FUNC.
        """
        if downsample_func is None:
            downsample_func = self._downsample_func_prefix
        # FLOATS
        if np.issubdtype(x_dtype, np.floating):
            if x_dtype == np.float16:
                return self._switch_mod_with_y(y_dtype, mod, f"{downsample_func}_f16")
            elif x_dtype == np.float32:
                return self._switch_mod_with_y(y_dtype, mod, f"{downsample_func}_f32")
            elif x_dtype == np.float64:
                return self._switch_mod_with_y(y_dtype, mod, f"{downsample_func}_f64")
        # UINTS
        elif np.issubdtype(x_dtype, np.unsignedinteger):
            if x_dtype == np.uint16:
                return self._switch_mod_with_y(y_dtype, mod, f"{downsample_func}_u16")
            elif x_dtype == np.uint32:
                return self._switch_mod_with_y(y_dtype, mod, f"{downsample_func}_u32")
            elif x_dtype == np.uint64:
                return self._switch_mod_with_y(y_dtype, mod, f"{downsample_func}_u64")
        # INTS (need to be last because uint is subdtype of int)
        elif np.issubdtype(x_dtype, np.integer):
            if x_dtype == np.int16:
                return self._switch_mod_with_y(y_dtype, mod, f"{downsample_func}_i16")
            elif x_dtype == np.int32:
                return self._switch_mod_with_y(y_dtype, mod, f"{downsample_func}_i32")
            elif x_dtype == np.int64:
                return self._switch_mod_with_y(y_dtype, mod, f"{downsample_func}_i64")
        # DATETIME -> i64 (datetime64 is viewed as int64)
        # TIMEDELTA -> i64 (timedelta64 is viewed as int64)
        raise ValueError(f"Unsupported data type (for x): {x_dtype}")

    def _downsample(
        self,
        x: Union[np.ndarray, None],
        y: np.ndarray,
        n_out: int,
        parallel: bool = False,
        **kwargs,
    ) -> np.ndarray:
        """Downsample the data in x and y."""
        mod = self.mod_single_core
        if parallel:
            if self.mod_multi_core is None:
                name = self.__class__.__name__
                warnings.warn(
                    f"No parallel implementation available for {name}. "
                    "Falling back to single-core implementation."
                )
            else:
                mod = self.mod_multi_core
        ## Viewing the y-data as different dtype (if necessary)
        y = self._view_y(y)
        ## Viewing the x-data as different dtype (if necessary)
        if x is None:
            downsample_f = self._switch_mod_with_y(y.dtype, mod)
            return downsample_f(y, n_out, **kwargs)
        x = self._view_x(x)
        ## Getting the appropriate downsample function
        downsample_f = self._switch_mod_with_x_and_y(x.dtype, y.dtype, mod)
        return downsample_f(x, y, n_out, **kwargs)

    def downsample(self, *args, n_out: int, parallel: bool = False, **kwargs):
        """Downsample the data in x and y.

        The x and y arguments are positional-only arguments. If only one argument is
        passed, it is considered to be the y-data. If two arguments are passed, the
        first argument is considered to be the x-data and the second argument is
        considered to be the y-data.
        """
        return super().downsample(*args, n_out=n_out, parallel=parallel, **kwargs)

    def __deepcopy__(self, memo):
        """Deepcopy the object."""
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        for k, v in self.__dict__.items():
            if k.endswith("_mod") or k.startswith("mod_"):
                # Don't (deep)copy the compiled modules
                setattr(result, k, v)
            else:
                setattr(result, k, deepcopy(v, memo))
        return result


NAN_DOWNSAMPLE_F = "downsample_nan"


class AbstractRustNaNDownsampler(AbstractRustDownsampler, ABC):
    """RustNaNDownsampler interface-class, subclassed by concrete downsamplers."""

    @property
    def _downsample_func_prefix(self) -> str:
        """The prefix of the downsample functions in the rust module."""
        return NAN_DOWNSAMPLE_F

    def _switch_mod_with_y(
        self, y_dtype: np.dtype, mod: ModuleType, downsample_func: Optional[str] = None
    ) -> Callable:
        """Select the appropriate function from the rust module for the y-data.

        Assumes equal binning (when no data for x is passed -> only this function is
        executed).
        Equidistant binning is  utilized when a `downsample_func` is passed from the
        `_switch_mod_with_x_and_y` method (since the x-data is considered in the
        downsampling).

        Parameters
        ----------
        y_dtype : np.dtype
            The dtype of the y-data
        mod : ModuleType
            The module to select the appropriate function from
        downsample_func : str, optional
            The name of the function to use, by default NAN_DOWNSAMPLE_F.
            This argument is passed from the `_switch_mod_with_x_and_y` method when
            the x-data is considered in the downsampling.
        """
        if downsample_func is None:
            downsample_func = self._downsample_func_prefix
        if not np.issubdtype(y_dtype, np.floating):
            # When y is not a float, we need to remove the _nan suffix to use the
            # regular downsample function as the _nan suffix is only used for floats.
            # (Note that NaNs only exist for floats)
            downsample_func = downsample_func.replace("_nan", "")
        return super()._switch_mod_with_y(y_dtype, mod, downsample_func)



import tsdownsample._rust._tsdownsample_rs as tsds_rs
from test_config import (
    rust_primitive_types_x,
    rust_primitive_types_y,
    rust_primitive_types_y_nan,
)


def _test_rust_mod_correctly_build(mod, sub_mods, has_x_impl: bool):
    # Without x
    for sub_mod in sub_mods:
        assert hasattr(mod, sub_mod)
        m = getattr(mod, sub_mod)
        for ty in rust_primitive_types_y:
            assert hasattr(m, f"downsample_{ty}")
    # With x
    if not has_x_impl:
        return
    for sub_mod in sub_mods:
        assert hasattr(mod, sub_mod)
        m = getattr(mod, sub_mod)
        for tx in rust_primitive_types_x:
            for ty in rust_primitive_types_y:
                assert hasattr(m, f"downsample_{tx}_{ty}")


def _test_rust_nan_mod_correctly_build(mod, sub_mods, has_x_impl: bool):
    # without x
    for sub_mod in sub_mods:
        assert hasattr(mod, sub_mod)
        m = getattr(mod, sub_mod)
        for ty in rust_primitive_types_y_nan:
            assert hasattr(m, f"downsample_nan_{ty}")

    # with x
    if not has_x_impl:
        return
    for sub_mod in sub_mods:
        assert hasattr(mod, sub_mod)
        m = getattr(mod, sub_mod)
        for tx in rust_primitive_types_x:
            for ty in rust_primitive_types_y_nan:
                assert hasattr(m, f"downsample_{tx}_{ty}")


def test_minmax_rust_mod_correctly_build():
    mod = tsds_rs.minmax
    sub_mods = ["sequential", "parallel"]
    _test_rust_mod_correctly_build(mod, sub_mods, has_x_impl=True)
    _test_rust_nan_mod_correctly_build(mod, sub_mods, has_x_impl=True)


def test_m4_rust_mod_correctly_build():
    mod = tsds_rs.m4
    sub_mods = ["sequential", "parallel"]
    _test_rust_mod_correctly_build(mod, sub_mods, has_x_impl=True)
    _test_rust_nan_mod_correctly_build(mod, sub_mods, has_x_impl=True)


def test_lttb_rust_mod_correctly_build():
    mod = tsds_rs.lttb
    sub_mods = ["sequential"]
    _test_rust_mod_correctly_build(mod, sub_mods, has_x_impl=True)


def test_minmaxlttb_rust_mod_correctly_build():
    mod = tsds_rs.minmaxlttb
    sub_mods = ["sequential", "parallel"]
    _test_rust_mod_correctly_build(mod, sub_mods, has_x_impl=True)
    _test_rust_nan_mod_correctly_build(mod, sub_mods, has_x_impl=True)



from typing import Iterable

import numpy as np
import pytest
from test_config import supported_dtypes_x, supported_dtypes_y

from tsdownsample import (  # MeanDownsampler,; MedianDownsampler,
    EveryNthDownsampler,
    LTTBDownsampler,
    M4Downsampler,
    MinMaxDownsampler,
    MinMaxLTTBDownsampler,
    NaNM4Downsampler,
    NaNMinMaxDownsampler,
    NaNMinMaxLTTBDownsampler,
)
from tsdownsample.downsampling_interface import (
    AbstractDownsampler,
    AbstractRustNaNDownsampler,
)

# TODO: Improve tests
#   - compare implementations with existing plotly_resampler implementations


RUST_DOWNSAMPLERS = [
    MinMaxDownsampler(),
    M4Downsampler(),
    LTTBDownsampler(),
    MinMaxLTTBDownsampler(),
]

RUST_NAN_DOWNSAMPLERS = [
    NaNMinMaxDownsampler(),
    NaNM4Downsampler(),
    NaNMinMaxLTTBDownsampler(),
]

OTHER_DOWNSAMPLERS = [EveryNthDownsampler()]


def generate_rust_downsamplers() -> Iterable[AbstractDownsampler]:
    for downsampler in RUST_DOWNSAMPLERS + RUST_NAN_DOWNSAMPLERS:
        yield downsampler


def generate_rust_nan_downsamplers() -> Iterable[AbstractDownsampler]:
    for downsampler in RUST_NAN_DOWNSAMPLERS:
        yield downsampler


def generate_all_downsamplers() -> Iterable[AbstractDownsampler]:
    for downsampler in RUST_DOWNSAMPLERS + RUST_NAN_DOWNSAMPLERS + OTHER_DOWNSAMPLERS:
        yield downsampler


def generate_datapoints():
    N_DATAPOINTS = 10_000
    return np.arange(N_DATAPOINTS)


def generate_nan_datapoints():
    N_DATAPOINTS = 10_000
    datapoints = np.arange(N_DATAPOINTS, dtype=np.float64)
    datapoints[0] = np.nan
    datapoints[9960] = np.nan
    return datapoints


@pytest.mark.parametrize("downsampler", generate_all_downsamplers())
def test_serialization_copy(downsampler: AbstractDownsampler):
    """Test serialization."""
    from copy import copy, deepcopy

    dc = copy(downsampler)
    ddc = deepcopy(downsampler)

    arr = generate_datapoints()

    orig_downsampled = downsampler.downsample(arr, n_out=100)
    dc_downsampled = dc.downsample(arr, n_out=100)
    ddc_downsampled = ddc.downsample(arr, n_out=100)
    assert np.all(orig_downsampled == dc_downsampled)
    assert np.all(orig_downsampled == ddc_downsampled)


@pytest.mark.parametrize("downsampler", generate_all_downsamplers())
def test_serialization_pickle(downsampler: AbstractDownsampler):
    """Test serialization."""
    import pickle

    dc = pickle.loads(pickle.dumps(downsampler))

    arr = generate_datapoints()
    orig_downsampled = downsampler.downsample(arr, n_out=100)
    dc_downsampled = dc.downsample(arr, n_out=100)
    assert np.all(orig_downsampled == dc_downsampled)


@pytest.mark.parametrize("downsampler", generate_rust_downsamplers())
def test_rust_downsampler(downsampler: AbstractDownsampler):
    """Test the Rust downsamplers."""
    arr = generate_datapoints()
    s_downsampled = downsampler.downsample(arr, n_out=100)
    assert s_downsampled[0] == 0
    assert s_downsampled[-1] == len(arr) - 1


@pytest.mark.parametrize("downsampler", generate_rust_nan_downsamplers())
def test_rust_nan_downsampler(downsampler: AbstractRustNaNDownsampler):
    """Test the Rust NaN downsamplers."""
    datapoints = generate_nan_datapoints()
    s_downsampled = downsampler.downsample(datapoints, n_out=100)
    print(s_downsampled)
    assert s_downsampled[0] == 0
    assert s_downsampled[-2] == 9960
    assert s_downsampled[50] != np.nan


def test_everynth_downsampler():
    """Test EveryNth downsampler."""
    arr = np.arange(10_000)
    downsampler = EveryNthDownsampler()
    s_downsampled = downsampler.downsample(arr, n_out=100)
    assert s_downsampled[0] == 0
    assert s_downsampled[-1] == 9_900


@pytest.mark.parametrize("downsampler", generate_rust_downsamplers())
def test_parallel_downsampling(downsampler: AbstractDownsampler):
    """Test parallel downsampling."""
    arr = np.random.randn(10_000).astype(np.float32)
    s_downsampled = downsampler.downsample(arr, n_out=100)
    s_downsampled_p = downsampler.downsample(arr, n_out=100, parallel=True)
    assert np.all(s_downsampled == s_downsampled_p)


@pytest.mark.parametrize("downsampler", generate_rust_downsamplers())
def test_parallel_downsampling_with_x(downsampler: AbstractDownsampler):
    """Test parallel downsampling with x."""
    arr = np.random.randn(10_001).astype(np.float32)  # 10_001 to test edge case
    idx = np.arange(len(arr))
    s_downsampled = downsampler.downsample(idx, arr, n_out=100)
    s_downsampled_p = downsampler.downsample(idx, arr, n_out=100, parallel=True)
    assert np.all(s_downsampled == s_downsampled_p)


@pytest.mark.parametrize("downsampler", generate_all_downsamplers())
def test_downsampling_with_x(downsampler: AbstractDownsampler):
    """Test downsampling with x."""
    arr = np.random.randn(2_001).astype(np.float32)  # 2_001 to test edge case
    idx = np.arange(len(arr))
    s_downsampled = downsampler.downsample(arr, n_out=100)
    s_downsampled_x = downsampler.downsample(idx, arr, n_out=100)
    assert np.all(s_downsampled == s_downsampled_x)


@pytest.mark.parametrize("downsampler", generate_all_downsamplers())
def test_downsampling_with_gaps_in_x(downsampler: AbstractDownsampler):
    """Test downsampling with gaps in x.

    With gap we do NOT mean a NaN in the array, but a large gap in the x values.
    """
    # TODO: might improve this test, now we just validate that the code does
    # not crash
    arr = np.random.randn(10_000).astype(np.float32)
    idx = np.arange(len(arr))
    idx[: len(idx) // 2] += len(idx) // 2  # add large gap in x
    s_downsampled = downsampler.downsample(idx, arr, n_out=100)
    assert len(s_downsampled) <= 100
    assert len(s_downsampled) >= 66


@pytest.mark.parametrize("downsampler", generate_rust_downsamplers())
def test_downsampling_different_dtypes(downsampler: AbstractDownsampler):
    """Test downsampling with different data types."""
    arr_orig = np.random.randint(0, 100, size=10_000)
    res = []
    for dtype_y in supported_dtypes_y:
        arr = arr_orig.astype(dtype_y)
        s_downsampled = downsampler.downsample(arr, n_out=100)
        if dtype_y is not np.bool_:
            res += [s_downsampled]
    for i in range(1, len(res)):
        assert np.all(res[0] == res[i])


@pytest.mark.parametrize("downsampler", generate_rust_downsamplers())
def test_downsampling_different_dtypes_with_x(downsampler: AbstractDownsampler):
    """Test downsampling with x with different data types."""
    arr_orig = np.random.randint(0, 100, size=10_000)
    idx_orig = np.arange(len(arr_orig))
    for dtype_x in supported_dtypes_x:
        res = []
        idx = idx_orig.astype(dtype_x)
        for dtype_y in supported_dtypes_y:
            arr = arr_orig.astype(dtype_y)
            s_downsampled = downsampler.downsample(idx, arr, n_out=100)
            if dtype_y is not np.bool_:
                res += [s_downsampled]
        for i in range(1, len(res)):
            assert np.all(res[0] == res[i])


@pytest.mark.parametrize("downsampler", generate_rust_downsamplers())
def test_downsampling_no_out_of_bounds_different_dtypes(
    downsampler: AbstractDownsampler,
):
    """Test no out of bounds issues when downsampling with different data types."""
    arr_orig = np.random.randint(0, 100, size=100)
    res = []
    for dtype in supported_dtypes_y:
        arr = arr_orig.astype(dtype)
        s_downsampled = downsampler.downsample(arr, n_out=76)
        s_downsampled_p = downsampler.downsample(arr, n_out=76, parallel=True)
        assert np.all(s_downsampled == s_downsampled_p)
        if dtype is not np.bool_:
            res += [s_downsampled]
    for i in range(1, len(res)):
        assert np.all(res[0] == res[i])


@pytest.mark.parametrize("downsampler", generate_rust_downsamplers())
def test_downsampling_no_out_of_bounds_different_dtypes_with_x(
    downsampler: AbstractDownsampler,
):
    """Test no out of bounds issues when downsampling with different data types."""
    arr_orig = np.random.randint(0, 100, size=100)
    idx_orig = np.arange(len(arr_orig))
    for dtype_x in supported_dtypes_x:
        res = []
        idx = idx_orig.astype(dtype_x)
        for dtype_y in supported_dtypes_y:
            arr = arr_orig.astype(dtype_y)
            s_downsampled = downsampler.downsample(idx, arr, n_out=76)
            s_downsampled_p = downsampler.downsample(idx, arr, n_out=76, parallel=True)
            assert np.all(s_downsampled == s_downsampled_p)
            if dtype_y is not np.bool_:
                res += [s_downsampled]
        for i in range(1, len(res)):
            assert np.all(res[0] == res[i])


def test_lttb_no_overflow():
    """Test no overflow when calculating average."""
    ### THIS SHOULD NOT OVERFLOW & HAVE THE SAME RESULT
    arr_orig = np.array([2 * 10**5] * 10_000, dtype=np.float64)
    s_downsampled = LTTBDownsampler().downsample(arr_orig, n_out=100)
    arr = arr_orig.astype(np.float32)
    s_downsampled_f32 = LTTBDownsampler().downsample(arr, n_out=100)
    assert np.all(s_downsampled == s_downsampled_f32)
    ### THIS SHOULD OVERFLOW & THUS HAVE A DIFFERENT RESULT...
    # max float32 is 3.4028235 × 1038 (so 2*10**38 is too big when adding 2 values)
    arr_orig = np.array([2 * 10**38] * 10_000, dtype=np.float64)
    s_downsampled = LTTBDownsampler().downsample(arr_orig, n_out=100)
    arr = arr_orig.astype(np.float32)
    s_downsampled_f32 = LTTBDownsampler().downsample(arr, n_out=100)
    assert not np.all(s_downsampled == s_downsampled_f32)  # TODO :(
    # I will leave this test here, but as many (much larger) libraries do not
    # really account for this, I guess it is perhaps less of an issue than I
    # thought. In the end f32 MAX is 3.4028235 × 1038 & f64 MAX is
    # 1.7976931348623157 × 10308 => which is in the end quite a lot.. (and all
    # integer averages are handled using f64) - f32 is only used for f16 & f32
    # (just as in numpy).


def test_invalid_nout():
    """Test invalid n_out."""
    arr = np.random.randint(0, 100, size=10_000)
    with pytest.raises(ValueError):
        LTTBDownsampler().downsample(arr, n_out=-1)
    with pytest.raises(ValueError):
        # Should be even
        MinMaxDownsampler().downsample(arr, n_out=33)
    with pytest.raises(ValueError):
        # Should be multiple of 4
        M4Downsampler().downsample(arr, n_out=34)


def test_error_unsupported_dtype():
    """Test unsupported dtype."""
    arr = np.random.randint(0, 100, size=10_000)
    arr = arr.astype("object")
    with pytest.raises(ValueError):
        MinMaxDownsampler().downsample(arr, n_out=100)


def test_error_invalid_args():
    """Test invalid arguments."""
    arr = np.random.randint(0, 100, size=10_000)
    # No args
    with pytest.raises(ValueError) as e_msg:
        MinMaxDownsampler().downsample(n_out=100, parallel=True)
    assert "takes 1 or 2 positional arguments" in str(e_msg.value)
    # Too many args
    with pytest.raises(ValueError) as e_msg:
        MinMaxDownsampler().downsample(arr, arr, arr, n_out=100, parallel=True)
    assert "takes 1 or 2 positional arguments" in str(e_msg.value)
    # Invalid y
    with pytest.raises(ValueError) as e_msg:
        MinMaxDownsampler().downsample(arr.reshape(5, 2_000), n_out=100, parallel=True)
    assert "y must be 1D" in str(e_msg.value)
    # Invalid x
    with pytest.raises(ValueError) as e_msg:
        MinMaxDownsampler().downsample(
            arr.reshape(5, 2_000), arr, n_out=100, parallel=True
        )
    assert "x must be 1D" in str(e_msg.value)
    # Invalid x and y (different length)
    with pytest.raises(ValueError) as e_msg:
        MinMaxDownsampler().downsample(arr, arr[:-1], n_out=100, parallel=True)
    assert "x and y must have the same length" in str(e_msg.value)


@pytest.mark.parametrize("downsampler", generate_rust_downsamplers())
def test_non_contiguous_array(downsampler: AbstractDownsampler):
    """Test non contiguous array."""
    arr = np.random.randint(0, 100, size=10_000).astype(np.float32)
    arr = arr[::2]
    assert not arr.flags["C_CONTIGUOUS"]
    with pytest.raises(ValueError) as e_msg:
        downsampler.downsample(arr, n_out=100)
    assert "must be contiguous" in str(e_msg.value)


def test_everynth_non_contiguous_array():
    """Test non contiguous array."""
    arr = np.random.randint(0, 100, size=10_000)
    arr = arr[::2]
    assert not arr.flags["C_CONTIGUOUS"]
    downsampler = EveryNthDownsampler()
    s_downsampled = downsampler.downsample(arr, n_out=100)
    assert s_downsampled[0] == 0
    assert s_downsampled[-1] == 4950


def test_nan_minmax_downsampler():
    """Test NaN downsamplers."""
    arr = np.random.randn(50_000)
    arr[::5] = np.nan
    s_downsampled = NaNMinMaxDownsampler().downsample(arr, n_out=100)
    arr_downsampled = arr[s_downsampled]
    assert np.all(np.isnan(arr_downsampled))


def test_nan_m4_downsampler():
    """Test NaN downsamplers."""
    arr = np.random.randn(50_000)
    arr[::5] = np.nan
    s_downsampled = NaNM4Downsampler().downsample(arr, n_out=100)
    arr_downsampled = arr[s_downsampled]
    assert np.all(np.isnan(arr_downsampled[1::4]))  # min is NaN
    assert np.all(np.isnan(arr_downsampled[2::4]))  # max is NaN


def test_nan_minmaxlttb_downsampler():
    """Test NaN downsamplers."""
    arr = np.random.randn(50_000)
    arr[::5] = np.nan
    s_downsampled = NaNMinMaxLTTBDownsampler().downsample(arr, n_out=100)
    arr_downsampled = arr[s_downsampled]
    assert np.all(np.isnan(arr_downsampled[1:-1]))  # first and last are not NaN



import numpy as np
import pytest

from tsdownsample import (
    EveryNthDownsampler,
    LTTBDownsampler,
    M4Downsampler,
    MinMaxDownsampler,
    MinMaxLTTBDownsampler,
    NaNM4Downsampler,
    NaNMinMaxDownsampler,
    NaNMinMaxLTTBDownsampler,
)

NB_SAMPLES = ["100,000", "1,000,000"]
N_OUT = ["100", "1,000", "5,000"]
Y_DTYPES = [np.float32, np.float64] + [np.int32, np.int64]


# --------------------------------------------------------------------------- #
#                               MinMaxDownsampler
# --------------------------------------------------------------------------- #


@pytest.mark.benchmark(group="minmax")
@pytest.mark.parametrize("n_samples", NB_SAMPLES)
@pytest.mark.parametrize("n_out", N_OUT)
@pytest.mark.parametrize("dtype", Y_DTYPES)
@pytest.mark.parametrize("parallel", [False, True])
def test_minmax_no_x(benchmark, n_samples, n_out, dtype, parallel):
    """Test the MinMaxDownsampler."""
    downsampler = MinMaxDownsampler()
    n_samples = int(n_samples.replace(",", ""))
    n_out = int(n_out.replace(",", ""))

    y = np.random.randn(n_samples).astype(dtype)

    benchmark(downsampler.downsample, y, n_out=n_out, parallel=parallel)


@pytest.mark.benchmark(group="minmax")
@pytest.mark.parametrize("n_samples", NB_SAMPLES)
@pytest.mark.parametrize("n_out", N_OUT)
@pytest.mark.parametrize("dtype", Y_DTYPES)
@pytest.mark.parametrize("parallel", [False, True])
def test_minmax_with_x(benchmark, n_samples, n_out, dtype, parallel):
    """Test the MinMaxDownsampler."""
    downsampler = MinMaxDownsampler()
    n_samples = int(n_samples.replace(",", ""))
    n_out = int(n_out.replace(",", ""))

    x = np.arange(n_samples)
    y = np.random.randn(n_samples).astype(dtype)

    benchmark(downsampler.downsample, x, y, n_out=n_out, parallel=parallel)


@pytest.mark.benchmark(group="nanminmax")
@pytest.mark.parametrize("n_samples", NB_SAMPLES)
@pytest.mark.parametrize("n_out", N_OUT)
@pytest.mark.parametrize("dtype", Y_DTYPES)
@pytest.mark.parametrize("parallel", [False, True])
def test_nanminmax_no_x(benchmark, n_samples, n_out, dtype, parallel):
    """Test the MinMaxDownsampler."""
    downsampler = NaNMinMaxDownsampler()
    n_samples = int(n_samples.replace(",", ""))
    n_out = int(n_out.replace(",", ""))

    y = np.random.randn(n_samples).astype(dtype)

    benchmark(downsampler.downsample, y, n_out=n_out, parallel=parallel)


@pytest.mark.benchmark(group="nanminmax")
@pytest.mark.parametrize("n_samples", NB_SAMPLES)
@pytest.mark.parametrize("n_out", N_OUT)
@pytest.mark.parametrize("dtype", Y_DTYPES)
@pytest.mark.parametrize("parallel", [False, True])
def test_nanminmax_with_x(benchmark, n_samples, n_out, dtype, parallel):
    """Test the MinMaxDownsampler."""
    downsampler = NaNMinMaxDownsampler()
    n_samples = int(n_samples.replace(",", ""))
    n_out = int(n_out.replace(",", ""))

    x = np.arange(n_samples)
    y = np.random.randn(n_samples).astype(dtype)

    benchmark(downsampler.downsample, x, y, n_out=n_out, parallel=parallel)


# --------------------------------------------------------------------------- #
#                               M4Downsampler
# --------------------------------------------------------------------------- #


@pytest.mark.benchmark(group="m4")
@pytest.mark.parametrize("n_samples", NB_SAMPLES)
@pytest.mark.parametrize("n_out", N_OUT)
@pytest.mark.parametrize("dtype", Y_DTYPES)
@pytest.mark.parametrize("parallel", [False, True])
def test_m4_no_x(benchmark, n_samples, n_out, dtype, parallel):
    """Test the M4Downsampler."""
    downsampler = M4Downsampler()
    n_samples = int(n_samples.replace(",", ""))
    n_out = int(n_out.replace(",", ""))

    y = np.random.randn(n_samples).astype(dtype)

    benchmark(downsampler.downsample, y, n_out=n_out, parallel=parallel)


@pytest.mark.benchmark(group="m4")
@pytest.mark.parametrize("n_samples", NB_SAMPLES)
@pytest.mark.parametrize("n_out", N_OUT)
@pytest.mark.parametrize("dtype", Y_DTYPES)
@pytest.mark.parametrize("parallel", [False, True])
def test_m4_with_x(benchmark, n_samples, n_out, dtype, parallel):
    """Test the M4Downsampler."""
    downsampler = M4Downsampler()
    n_samples = int(n_samples.replace(",", ""))
    n_out = int(n_out.replace(",", ""))

    x = np.arange(n_samples)
    y = np.random.randn(n_samples).astype(dtype)

    benchmark(downsampler.downsample, x, y, n_out=n_out, parallel=parallel)


@pytest.mark.benchmark(group="nanm4")
@pytest.mark.parametrize("n_samples", NB_SAMPLES)
@pytest.mark.parametrize("n_out", N_OUT)
@pytest.mark.parametrize("dtype", Y_DTYPES)
@pytest.mark.parametrize("parallel", [False, True])
def test_nanm4_no_x(benchmark, n_samples, n_out, dtype, parallel):
    """Test the M4Downsampler."""
    downsampler = NaNM4Downsampler()
    n_samples = int(n_samples.replace(",", ""))
    n_out = int(n_out.replace(",", ""))

    y = np.random.randn(n_samples).astype(dtype)

    benchmark(downsampler.downsample, y, n_out=n_out, parallel=parallel)


@pytest.mark.benchmark(group="nanm4")
@pytest.mark.parametrize("n_samples", NB_SAMPLES)
@pytest.mark.parametrize("n_out", N_OUT)
@pytest.mark.parametrize("dtype", Y_DTYPES)
@pytest.mark.parametrize("parallel", [False, True])
def test_nanm4_with_x(benchmark, n_samples, n_out, dtype, parallel):
    """Test the M4Downsampler."""
    downsampler = NaNM4Downsampler()
    n_samples = int(n_samples.replace(",", ""))
    n_out = int(n_out.replace(",", ""))

    x = np.arange(n_samples)
    y = np.random.randn(n_samples).astype(dtype)

    benchmark(downsampler.downsample, x, y, n_out=n_out, parallel=parallel)


# --------------------------------------------------------------------------- #
#                              LTTBDownsampler
# --------------------------------------------------------------------------- #


@pytest.mark.benchmark(group="lttb")
@pytest.mark.parametrize("n_samples", NB_SAMPLES)
@pytest.mark.parametrize("n_out", N_OUT)
@pytest.mark.parametrize("dtype", Y_DTYPES)
@pytest.mark.parametrize("parallel", [False, True])
def test_lttb_no_x(benchmark, n_samples, n_out, dtype, parallel):
    """Test the LTTBDownsampler."""
    downsampler = LTTBDownsampler()
    n_samples = int(n_samples.replace(",", ""))
    n_out = int(n_out.replace(",", ""))

    y = np.random.randn(n_samples).astype(dtype)

    benchmark(downsampler.downsample, y, n_out=n_out, parallel=parallel)


@pytest.mark.benchmark(group="lttb")
@pytest.mark.parametrize("n_samples", NB_SAMPLES)
@pytest.mark.parametrize("n_out", N_OUT)
@pytest.mark.parametrize("dtype", Y_DTYPES)
@pytest.mark.parametrize("parallel", [False, True])
def test_lttb_with_x(benchmark, n_samples, n_out, dtype, parallel):
    """Test the LTTBDownsampler."""
    downsampler = LTTBDownsampler()
    n_samples = int(n_samples.replace(",", ""))
    n_out = int(n_out.replace(",", ""))

    x = np.arange(n_samples)
    y = np.random.randn(n_samples).astype(dtype)

    benchmark(downsampler.downsample, x, y, n_out=n_out, parallel=parallel)


# --------------------------------------------------------------------------- #
#                          MinMaxLTTBDownsampler
# --------------------------------------------------------------------------- #


@pytest.mark.benchmark(group="minmaxlttb")
@pytest.mark.parametrize("n_samples", NB_SAMPLES)
@pytest.mark.parametrize("n_out", N_OUT)
@pytest.mark.parametrize("dtype", Y_DTYPES)
@pytest.mark.parametrize("parallel", [False, True])
def test_minmaxlttb_no_x(benchmark, n_samples, n_out, dtype, parallel):
    """Test the MinMaxLTTBDownsampler."""
    downsampler = MinMaxLTTBDownsampler()
    n_samples = int(n_samples.replace(",", ""))
    n_out = int(n_out.replace(",", ""))

    y = np.random.randn(n_samples).astype(dtype)

    benchmark(downsampler.downsample, y, n_out=n_out, parallel=parallel)


@pytest.mark.benchmark(group="minmaxlttb")
@pytest.mark.parametrize("n_samples", NB_SAMPLES)
@pytest.mark.parametrize("n_out", N_OUT)
@pytest.mark.parametrize("dtype", Y_DTYPES)
@pytest.mark.parametrize("parallel", [False, True])
def test_minmaxlttb_with_x(benchmark, n_samples, n_out, dtype, parallel):
    """Test the MinMaxLTTBDownsampler."""
    downsampler = MinMaxLTTBDownsampler()
    n_samples = int(n_samples.replace(",", ""))
    n_out = int(n_out.replace(",", ""))

    x = np.arange(n_samples)
    y = np.random.randn(n_samples).astype(dtype)

    benchmark(downsampler.downsample, x, y, n_out=n_out, parallel=parallel)


@pytest.mark.benchmark(group="nanminmaxlttb")
@pytest.mark.parametrize("n_samples", NB_SAMPLES)
@pytest.mark.parametrize("n_out", N_OUT)
@pytest.mark.parametrize("dtype", Y_DTYPES)
@pytest.mark.parametrize("parallel", [False, True])
def test_nanminmaxlttb_no_x(benchmark, n_samples, n_out, dtype, parallel):
    """Test the MinMaxLTTBDownsampler."""
    downsampler = NaNMinMaxLTTBDownsampler()
    n_samples = int(n_samples.replace(",", ""))
    n_out = int(n_out.replace(",", ""))

    y = np.random.randn(n_samples).astype(dtype)

    benchmark(downsampler.downsample, y, n_out=n_out, parallel=parallel)


@pytest.mark.benchmark(group="nanminmaxlttb")
@pytest.mark.parametrize("n_samples", NB_SAMPLES)
@pytest.mark.parametrize("n_out", N_OUT)
@pytest.mark.parametrize("dtype", Y_DTYPES)
@pytest.mark.parametrize("parallel", [False, True])
def test_nanminmaxlttb_with_x(benchmark, n_samples, n_out, dtype, parallel):
    """Test the MinMaxLTTBDownsampler."""
    downsampler = NaNMinMaxLTTBDownsampler()
    n_samples = int(n_samples.replace(",", ""))
    n_out = int(n_out.replace(",", ""))

    x = np.arange(n_samples)
    y = np.random.randn(n_samples).astype(dtype)

    benchmark(downsampler.downsample, x, y, n_out=n_out, parallel=parallel)


# --------------------------------------------------------------------------- #
#                             EveryNthDownsampler
# --------------------------------------------------------------------------- #


@pytest.mark.benchmark(group="everynth")
@pytest.mark.parametrize("n_samples", NB_SAMPLES)
@pytest.mark.parametrize("n_out", N_OUT)
def test_everynth(benchmark, n_samples, n_out):
    """Test the EveryNthDownsampler."""
    downsampler = EveryNthDownsampler()
    n_samples = int(n_samples.replace(",", ""))
    n_out = int(n_out.replace(",", ""))

    y = np.random.randn(n_samples)

    benchmark(downsampler.downsample, y, n_out=n_out)
